/*
 * GPL HEADER START
 *
 * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 only,
 * as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License version 2 for more details (a copy is included
 * in the LICENSE file that accompanied this code).
 *
 * You should have received a copy of the GNU General Public License
 * version 2 along with this program; If not, see http://www.gnu.org/licenses
 *
 * GPL HEADER END
 */

/*
 * Copyright (c) 2016 DDN Storage
 * Author: Sebastien Buisson sbuisson@ddn.com
 */

/*
 * lustre/utils/l_getsepol.c
 * Userland helper to retrieve SELinux policy information.
 */

#include <sys/types.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <libgen.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <syslog.h>
#include <stdarg.h>
#include <fcntl.h>
#include <stddef.h>
#include <ctype.h>
#include <dirent.h>

#include <openssl/evp.h>

#include <selinux/selinux.h>

#include <libcfs/util/param.h>
#include <lustre/lustre_user.h>
#include <lustre/lustre_idl.h>


static char *progname;

static void errlog(const char *fmt, ...)
{
	va_list args;

	openlog(progname, LOG_PERROR | LOG_PID, LOG_AUTHPRIV);

	va_start(args, fmt);
	vsyslog(LOG_NOTICE, fmt, args);
	va_end(args);

	closelog();
}

/* Read binary SELinux policy, and compute hash */
static int sepol_get_policy_data(unsigned char **mdval, unsigned int *mdsize,
				 int *policyver)
{
	int fd;
	char entry_path[100];
	glob_t glob_info = { 0 };
	char buffer[1024];
	ssize_t count = 1024;
	EVP_MD_CTX *mdctx;
	const EVP_MD *md = EVP_sha256(); /* use SHA-256 */
	char *polver;
	int rc;

	snprintf(entry_path, sizeof(entry_path), "%s.*",
		 selinux_binary_policy_path());

	/* Find policy file */
	rc = glob(entry_path, GLOB_ERR, NULL, &glob_info);
	if (rc) {
		errlog("can't find SELinux policy file %s.*: %d\n",
		       entry_path, rc);
		goto out;
	}
	if (glob_info.gl_pathc != 1) {
		errlog("can't choose between %d SELinux policy files %s.*\n",
		       glob_info.gl_pathc, entry_path);
		rc = -ENOENT;
		goto out;
	}

	/* Get policy version from policy file name */
	polver = strrchr(glob_info.gl_pathv[0], '.');
	if (polver == NULL) {
		rc = -ENOENT;
		goto out;
	}
	*policyver = atoi(polver + 1);

	/* Open policy file */
	fd = open(glob_info.gl_pathv[0], O_RDONLY);
	if (fd < 0) {
		errlog("can't open SELinux policy file %s: %s\n", entry_path,
		       strerror(errno));
		rc = -ENOENT;
		goto out;
	}

	/* Read policy file */
	mdctx = EVP_MD_CTX_create();
	EVP_DigestInit_ex(mdctx, md, NULL);
	while (count == 1024) {
		count = read(fd, buffer, count);
		if (count < 0) {
			errlog("can't read SELinux policy file %s\n",
			       entry_path);
			rc = -errno;
			close(fd);
			goto out;
		}
		EVP_DigestUpdate(mdctx, buffer, count);
	}

	/* Close policy file */
	rc = close(fd);
	if (rc < 0) {
		rc = -errno;
		goto out;
	}

	*mdsize = EVP_MD_size(md);
	*mdval = malloc(*mdsize);
	if (*mdval == NULL) {
		rc = -ENOMEM;
		goto out;
	}

	EVP_DigestFinal_ex(mdctx, *mdval, NULL);
	EVP_MD_CTX_destroy(mdctx);

out:
	globfree(&glob_info);
	return rc;
}

/**
 * Calculate SELinux status information.
 * String that represents SELinux status info has the following format:
 * <mode>:<policy name>:<policy version>:<policy hash>
 * <mode> is a digit equal to 0 for SELinux Permissive mode,
 * and 1 for Enforcing mode.
 * When called from kernel space, it requires 4 args:
 * - obd type
 * - obd name
 * - pointer to request
 * - pointer to import
 * When called without proper args, it prints calculated info
 * to stdout.
 */
int main(int argc, char **argv)
{
	struct sepol_downcall_data *data = NULL;
	unsigned long ptr;
	glob_t path;
	int fd, size;
	int enforce;
	char *policy_type = NULL;
	int policyver = 0;
	unsigned char *mdval = NULL;
	unsigned int mdsize = 0;
	char *p;
	int idx;
	int rc = 0;

	progname = basename(argv[0]);

	size = offsetof(struct sepol_downcall_data,
			sdd_sepol[LUSTRE_NODEMAP_SEPOL_LENGTH + 1]);
	data = malloc(size);
	if (!data) {
		errlog("malloc sepol downcall data(%d) failed!\n", size);
		rc = -ENOMEM;
		goto out;
	}

	memset(data, 0, size);
	data->sdd_magic = SEPOL_DOWNCALL_MAGIC;

	/* Determine if SELinux is in permissive or enforcing mode */
	enforce = security_getenforce();
	rc = enforce;
	if (rc < 0)
		goto out;

	/* Get policy name */
	rc = selinux_getpolicytype(&policy_type);
	if (rc < 0)
		goto out;

	/* Read binary SELinux policy, and compute hash */
	rc = sepol_get_policy_data(&mdval, &mdsize, &policyver);
	if (rc < 0)
		goto out_poltyp;

	/* Put all info together and generate string
	 * to represent SELinux policy information */
	rc = snprintf(data->sdd_sepol, LUSTRE_NODEMAP_SEPOL_LENGTH + 1,
		      "%.1d:%s:%u:", enforce, policy_type, policyver);
	if (rc >= LUSTRE_NODEMAP_SEPOL_LENGTH + 1) {
		rc = -EMSGSIZE;
		goto out_mdval;
	}

	p = data->sdd_sepol + strlen(data->sdd_sepol);
	size = LUSTRE_NODEMAP_SEPOL_LENGTH + 1 - strlen(data->sdd_sepol);
	for (idx = 0; idx < mdsize; idx++) {
		rc = snprintf(p, size, "%02x",
			      (unsigned char)(mdval[idx]));
		p += 2;
		size -= 2;
		if (size < 0 || rc >= size) {
			rc = -EMSGSIZE;
			goto out_mdval;
		}
	}
	data->sdd_sepol_len = p - data->sdd_sepol;

	size = offsetof(struct sepol_downcall_data,
			sdd_sepol[data->sdd_sepol_len]);

	if (argc < 5) {
		if (size == 0 || data->sdd_sepol_len == 0)
			errlog("failed to get SELinux status info\n");
		printf("SELinux status info: %.*s\n",
		       data->sdd_sepol_len, data->sdd_sepol);
		return rc;
	}

	ptr = strtoul(argv[3], NULL, 16);
	data->sdd_req = ptr;
	ptr = strtoul(argv[4], NULL, 16);
	data->sdd_imp = ptr;

	/* Send SELinux policy info to kernelspace */
	rc = cfs_get_param_paths(&path, "%s/%s/srpc_sepol", argv[1], argv[2]);
	if (rc != 0) {
		rc = -errno;
		goto out;
	}

	fd = open(path.gl_pathv[0], O_WRONLY);
	if (fd < 0) {
		errlog("can't open file '%s':%s\n", path.gl_pathv[0],
		       strerror(errno));
		rc = -errno;
		goto out_params;
	}

	rc = write(fd, data, size);
	close(fd);
	if (rc != size) {
		errlog("partial write ret %d: %s\n", rc, strerror(errno));
		rc = -errno;
	} else {
		rc = 0;
	}

out_params:
	cfs_free_param_data(&path);
out_mdval:
	free(mdval);
out_poltyp:
	free(policy_type);
out:
	free(data);

	return rc;
}
